import itertools
from queue import Queue
from typing import Optional, Set

import gavel.logic.logic as logic


# basic functionality for manipulating FOL formulas


def binary_to_nary(
    formula: logic.LogicExpression, connective: logic.BinaryConnective
) -> logic.NaryFormula:
    assert connective.is_associative()
    if isinstance(formula, logic.BinaryFormula) and formula.operator == connective:
        return logic.NaryFormula(
            connective,
            list(itertools.chain(
                binary_to_nary(formula.left, connective).formulae,
                binary_to_nary(formula.right, connective).formulae,
            )),
        )
    if isinstance(formula, logic.NaryFormula) and formula.operator == connective:
        return logic.NaryFormula(
            connective,
            [
                f
                for sub in formula.formulae
                for f in binary_to_nary(sub, connective).formulae
            ],
        )
    else:
        return logic.NaryFormula(connective, [formula])


def get_vars_in_formula(formula: logic.LogicElement) -> Set[logic.Variable]:
    if isinstance(formula, logic.Variable):
        return {formula}
    elif isinstance(formula, logic.PredicateExpression):
        return set(arg for arg in formula.arguments if isinstance(arg, logic.Variable))
    elif isinstance(formula, logic.BinaryFormula):
        return get_vars_in_formula(formula.left).union(
            get_vars_in_formula(formula.right)
        )
    elif isinstance(formula, logic.UnaryFormula) or isinstance(
        formula, logic.QuantifiedFormula
    ):
        return get_vars_in_formula(formula.formula)
    elif isinstance(formula, logic.NaryFormula):
        variables = [get_vars_in_formula(f) for f in formula.formulae]
        variables = [v for v in variables if v is not None]
        if len(variables) > 0:
            try:
                return set.union(*variables)
            except TypeError:
                return set()
    else:
        return set()


def substitute_var_in_formula(
    formula: logic.LogicElement, var: Optional[logic.Variable], ind
):
    """Replace every occurrence of variable var with ind. If var is None, replace all variables with ind"""
    if isinstance(formula, logic.NaryFormula):
        return logic.NaryFormula(
            formula.operator,
            [substitute_var_in_formula(f, var, ind) for f in formula.formulae],
        )
    elif isinstance(formula, logic.PredicateExpression):
        return logic.PredicateExpression(
            formula.predicate,
            [
                (
                    ind
                    if isinstance(arg, logic.Variable) and (not var or arg == var)
                    else arg
                )
                for arg in formula.arguments
            ],
        )
    elif isinstance(formula, logic.UnaryFormula):
        return logic.UnaryFormula(
            formula.connective, substitute_var_in_formula(formula.formula, var, ind)
        )
    elif isinstance(formula, logic.QuantifiedFormula):
        variables = [arg for arg in formula.variables if arg != var]
        return logic.QuantifiedFormula(
            formula.quantifier,
            variables,
            substitute_var_in_formula(formula.formula, var, ind),
        )
    elif isinstance(formula, logic.BinaryFormula):
        left = substitute_var_in_formula(formula.left, var, ind)
        right = substitute_var_in_formula(formula.right, var, ind)
        return logic.BinaryFormula(left, formula.operator, right)
    elif isinstance(formula, logic.Variable):
        return ind if not var or formula == var else formula
    return formula


def replace_predicate_symbol(formula: logic.LogicElement, old: str, new: str):
    """Replace every occurrence of a predicate symbol in a given formula"""
    if isinstance(formula, logic.PredicateExpression) and formula.predicate == old:
        formula.predicate = new
    elif isinstance(formula, logic.UnaryFormula):
        formula.formula = replace_predicate_symbol(formula.formula, old, new)
    elif isinstance(formula, logic.QuantifiedFormula):
        formula.formula = replace_predicate_symbol(formula.formula, old, new)
    elif isinstance(formula, logic.BinaryFormula):
        formula.left = replace_predicate_symbol(formula.left, old, new)
        formula.right = replace_predicate_symbol(formula.right, old, new)
    elif isinstance(formula, logic.NaryFormula):
        formula.formulae = [
            replace_predicate_symbol(f, old, new) for f in formula.formulae
        ]
    return formula


def is_atom_formula(formula: logic.LogicExpression):
    if not isinstance(formula, logic.PredicateExpression):
        return False
    # allow integers as terms
    return isinstance(formula.arguments, logic.TermExpression) or all(
        isinstance(t, logic.TermExpression) or isinstance(t, int)
        for t in formula.arguments
    )


def is_literal(formula: logic.LogicExpression):
    return (
        is_atom_formula(formula)
        or (
            isinstance(formula, logic.UnaryFormula) and is_atom_formula(formula.formula)
        )
        or (
            isinstance(formula, logic.BinaryFormula)
            and (
                formula.operator == logic.BinaryConnective.EQ
                or formula.operator == logic.BinaryConnective.NEQ
            )
        )
    )


def convert_biimplication_to_cnf(formula: logic.BinaryFormula):
    """a <-> b <=> (~a | b) & (~b | a)"""
    assert formula.operator == logic.BinaryConnective.BIIMPLICATION
    return logic.BinaryFormula(
        logic.BinaryFormula(
            logic.UnaryFormula(logic.UnaryConnective.NEGATION, formula.left),
            logic.BinaryConnective.DISJUNCTION,
            formula.right,
        ),
        logic.BinaryConnective.CONJUNCTION,
        logic.BinaryFormula(
            formula.left,
            logic.BinaryConnective.DISJUNCTION,
            logic.UnaryFormula(logic.UnaryConnective.NEGATION, formula.right),
        ),
    )


def convert_implication_to_disjunction(formula: logic.BinaryFormula):
    """a -> b <=> ~a | b"""
    assert formula.operator == logic.BinaryConnective.IMPLICATION
    return logic.BinaryFormula(
        logic.UnaryFormula(logic.UnaryConnective.NEGATION, formula.left),
        logic.BinaryConnective.DISJUNCTION,
        formula.right,
    )


def convert_to_nnf(formula: logic.LogicExpression):
    """Replace all <-> with (~a|b)&(~b|a), all -> with (~a|b), all ~(a!=b) with a=b, all ~(a=b) with a!=b,
    all ~(a|b) with (~a&~b), all ~(a&b) with (~a|~b), all ~![] with ?[]~, all ~?[] with ?[]~, all ~~a with a.
    Resulting formula will only have |, & and ~ operations and ~ only for literals"""
    if (
        isinstance(formula, logic.UnaryFormula)
        and formula.connective == logic.UnaryConnective.NEGATION
    ):
        if isinstance(formula.formula, logic.QuantifiedFormula):
            if formula.formula.quantifier == logic.Quantifier.EXISTENTIAL:
                quantifier = logic.Quantifier.UNIVERSAL
            else:
                quantifier = logic.Quantifier.EXISTENTIAL
            return logic.QuantifiedFormula(
                quantifier,
                formula.formula.variables,
                convert_to_nnf(
                    logic.UnaryFormula(
                        logic.UnaryConnective.NEGATION, formula.formula.formula
                    )
                ),
            )
        elif (
            isinstance(formula.formula, logic.BinaryFormula)
            and formula.formula.operator == logic.BinaryConnective.BIIMPLICATION
        ):
            return convert_to_nnf(
                logic.UnaryFormula(
                    logic.UnaryConnective.NEGATION,
                    convert_biimplication_to_cnf(formula.formula),
                )
            )
        elif (
            isinstance(formula.formula, logic.BinaryFormula)
            and formula.formula.operator == logic.BinaryConnective.IMPLICATION
        ):
            return convert_to_nnf(
                logic.UnaryFormula(
                    logic.UnaryConnective.NEGATION,
                    convert_implication_to_disjunction(formula.formula),
                )
            )
        elif isinstance(formula.formula, logic.BinaryFormula) or isinstance(
            formula.formula, logic.NaryFormula
        ):
            if formula.formula.operator == logic.BinaryConnective.DISJUNCTION:
                operator = logic.BinaryConnective.CONJUNCTION
            elif formula.formula.operator == logic.BinaryConnective.CONJUNCTION:
                operator = logic.BinaryConnective.DISJUNCTION
            elif formula.formula.operator == logic.BinaryConnective.EQ:
                assert isinstance(formula.formula, logic.BinaryFormula)
                return logic.BinaryFormula(
                    formula.formula.left,
                    logic.BinaryConnective.NEQ,
                    formula.formula.right,
                )
            elif formula.formula.operator == logic.BinaryConnective.NEQ:
                assert isinstance(formula.formula, logic.BinaryFormula)
                return logic.BinaryFormula(
                    formula.formula.left,
                    logic.BinaryConnective.EQ,
                    formula.formula.right,
                )
            else:
                raise NotImplementedError
            if isinstance(formula.formula, logic.BinaryFormula):
                return logic.BinaryFormula(
                    convert_to_nnf(
                        logic.UnaryFormula(
                            logic.UnaryConnective.NEGATION, formula.formula.left
                        )
                    ),
                    operator,
                    convert_to_nnf(
                        logic.UnaryFormula(
                            logic.UnaryConnective.NEGATION, formula.formula.right
                        )
                    ),
                )
            else:
                return logic.NaryFormula(
                    operator,
                    [
                        convert_to_nnf(
                            logic.UnaryFormula(logic.UnaryConnective.NEGATION, f)
                        )
                        for f in formula.formula.formulae
                    ],
                )
        elif (
            isinstance(formula.formula, logic.UnaryFormula)
            and formula.formula.connective == logic.UnaryConnective.NEGATION
        ):
            return convert_to_nnf(formula.formula.formula)
        else:
            return formula
    else:
        if isinstance(formula, logic.QuantifiedFormula):
            return logic.QuantifiedFormula(
                formula.quantifier, formula.variables, convert_to_nnf(formula.formula)
            )
        elif (
            isinstance(formula, logic.BinaryFormula)
            and formula.operator == logic.BinaryConnective.BIIMPLICATION
        ):
            return convert_to_nnf(convert_biimplication_to_cnf(formula))
        elif (
            isinstance(formula, logic.BinaryFormula)
            and formula.operator == logic.BinaryConnective.IMPLICATION
        ):
            return convert_to_nnf(convert_implication_to_disjunction(formula))
        elif isinstance(formula, logic.BinaryFormula):
            return logic.BinaryFormula(
                convert_to_nnf(formula.left),
                formula.operator,
                convert_to_nnf(formula.right),
            )
        elif isinstance(formula, logic.NaryFormula):
            return logic.NaryFormula(
                formula.operator, [convert_to_nnf(f) for f in formula.formulae]
            )
        else:
            return formula


def convert_to_cnf(formula: logic.LogicExpression) -> logic.NaryFormula:
    """Assumes no quantifiers"""
    formula = convert_to_nnf(formula)
    clauses = []
    unprocessed_clauses = Queue()
    unprocessed_clauses.put(formula)

    while not unprocessed_clauses.empty():
        clause = unprocessed_clauses.get()
        if isinstance(clause, logic.NaryFormula) or isinstance(
            clause, logic.BinaryFormula
        ):
            # nested conjunctions: resolve
            if clause.operator == logic.BinaryConnective.CONJUNCTION:
                if isinstance(clause, logic.BinaryFormula):
                    unprocessed_clauses.put(clause.left)
                    unprocessed_clauses.put(clause.right)
                else:
                    assert isinstance(clause, logic.NaryFormula)
                    for f in clause.formulae:
                        unprocessed_clauses.put(f)
            # disjunctions: apply cnf recursively, then distribute
            # i.e. [(A|B)&(C|D)] | [(E|F)&(G|H)] -> (A|B|E|F)&(A|B|G|H)&(C|D|E|F)&(C|D|G|H)
            elif clause.operator == logic.BinaryConnective.DISJUNCTION:
                if isinstance(clause, logic.BinaryFormula):
                    disj_parts = [clause.left, clause.right]
                else:
                    assert isinstance(clause, logic.NaryFormula)
                    disj_parts = clause.formulae
                disj_parts = [convert_to_cnf(p).formulae for p in disj_parts]
                for combination in itertools.product(*disj_parts):
                    clause_elements = []
                    for c in combination:
                        clause_elements += c.formulae
                    clauses.append(
                        logic.NaryFormula(
                            logic.BinaryConnective.DISJUNCTION, clause_elements
                        )
                    )
            else:
                clauses.append(
                    logic.NaryFormula(logic.BinaryConnective.DISJUNCTION, [clause])
                )
        else:
            clauses.append(
                logic.NaryFormula(logic.BinaryConnective.DISJUNCTION, [clause])
            )

    return logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, clauses)



_stati = {}


def get_status(name):
    return _stati.get(name)


class Status:
    _name = None

    def __init_subclass__(cls, **kwargs):
        if cls._name:
            _stati[cls._name] = cls


class StatusSuccess(Status):
    _name = "Success"
    _abbreviation = "SUC"
    _description = "The logical data has been processed successfully."


class StatusUnsatisfiabilityPreserving(StatusSuccess):
    _name = "UnsatisfiabilityPreserving"
    _abbreviation = "UNP"
    _description = "C, i.e., if Ax is unsatisfiable then C is unsatisfiable."


class StatusSatisfiabilityPreserving(StatusSuccess):
    _name = "SatisfiabilityPreserving"
    _abbreviation = "SAP"
    _description = "- F is satisfiable."


class StatusEquiSatisfiable(
    StatusSatisfiabilityPreserving, StatusUnsatisfiabilityPreserving
):
    _name = "EquiSatisfiable"
    _abbreviation = "ESA"
    _description = "(un)satisfiable iff C is (un)satisfiable."


class StatusSatisfiable(StatusEquiSatisfiable):
    _name = "Satisfiable"
    _abbreviation = "SAT"
    _description = "- Possible dataforms are Models of Ax | C."


class StatusFinitelySatisfiable(StatusSatisfiable):
    _name = "FinitelySatisfiable"
    _abbreviation = "FSA"
    _description = "- Possible dataforms are FiniteModels of Ax | C."


class StatusFiniteTheorem(StatusSuccess):
    _name = "FiniteTheorem"
    _abbreviation = "FTH"
    _description = "- Any models of Ax | ~C are infinite."


class StatusTheorem(StatusSatisfiabilityPreserving, StatusFiniteTheorem):
    _name = "Theorem"
    _abbreviation = "THM"
    _description = "- Possible dataforms are Proofs of C from Ax."


class StatusSatisfiableTheorem(StatusTheorem):
    _name = "SatisfiableTheorem"
    _abbreviation = "STH"
    _description = "- Possible dataforms are Models of Ax with Proofs of C from Ax."


class StatusEquivalent(StatusSatisfiable, StatusSatisfiableTheorem):
    _name = "Equivalent"
    _abbreviation = "EQV"
    _description = "- Possible dataforms are Proofs of C from Ax and of Ax from C."


class StatusTautologousConclusion(StatusSatisfiable, StatusSatisfiableTheorem):
    _name = "TautologousConclusion"
    _abbreviation = "TAC"
    _description = "- Possible dataforms are Proofs of C."


class StatusWeakerConclusion(StatusSatisfiable, StatusSatisfiableTheorem):
    _name = "WeakerConclusion"
    _abbreviation = "WEC"
    _description = "- See Theorem and Satisfiable."


class StatusEquivalentTheorem(StatusEquivalent):
    _name = "EquivalentTheorem"
    _abbreviation = "ETH"
    _description = "- See Equivalent."


class StatusTautology(StatusEquivalent, StatusTautologousConclusion):
    _name = "Tautology"
    _abbreviation = "TAU"
    _description = "- Possible dataforms are Proofs of Ax and of C."


class StatusWeakerTautologousConclusion(
    StatusTautologousConclusion, StatusWeakerConclusion
):
    _name = "WeakerTautologousConclusion"
    _abbreviation = "WTC"
    _description = "- See TautologousConclusion and WeakerConclusion."


class StatusWeakerTheorem(StatusWeakerConclusion):
    _name = "WeakerTheorem"
    _abbreviation = "WTH"
    _description = "- See Theorem and Satisfiable."


class StatusCounterSatisfiabilityPreserving(StatusSuccess):
    _name = "CounterSatisfiabilityPreserving"
    _abbreviation = "CSP"
    _description = "is satisfiable then ~C is satisfiable."


class StatusCounterTheorem(StatusCounterSatisfiabilityPreserving):
    _name = "CounterTheorem"
    _abbreviation = "CTH"
    _description = "- Possible dataforms are Proofs of ~C from Ax."


class StatusContradictoryAxioms(StatusTheorem, StatusCounterTheorem):
    _name = "ContradictoryAxioms"
    _abbreviation = "CAX"
    _description = "- Possible dataforms are Refutations of Ax."


class StatusSatisfiableConclusionContradictoryAxioms(StatusContradictoryAxioms):
    _name = "SatisfiableConclusionContradictoryAxioms"
    _abbreviation = "SCA"
    _description = "- See ContradictoryAxioms."


class StatusTautologousConclusionContradictoryAxioms(
    StatusSatisfiableConclusionContradictoryAxioms
):
    _name = "TautologousConclusionContradictoryAxioms"
    _abbreviation = "TCA"
    _description = (
        "- See TautologousConclusion and SatisfiableConclusionContradictoryAxioms."
    )


class StatusSatisfiableCounterConclusionContradictoryAxioms(StatusContradictoryAxioms):
    _name = "SatisfiableCounterConclusionContradictoryAxioms"
    _abbreviation = "SCC"
    _description = "- See ContradictoryAxioms."


class StatusWeakerConclusionContradictoryAxioms(
    StatusSatisfiableConclusionContradictoryAxioms,
    StatusSatisfiableCounterConclusionContradictoryAxioms,
):
    _name = "WeakerConclusionContradictoryAxioms"
    _abbreviation = "WCA"
    _description = "SatisfiableCounterConclusionContradictoryAxioms."


class StatusCounterUnsatisfiabilityPreserving(StatusSuccess):
    _name = "CounterUnsatisfiabilityPreserving"
    _abbreviation = "CUP"
    _description = "~C, i.e., if Ax is unsatisfiable then ~C is unsatisfiable."


class StatusEquiCounterSatisfiable(
    StatusCounterUnsatisfiabilityPreserving, StatusCounterSatisfiabilityPreserving
):
    _name = "EquiCounterSatisfiable"
    _abbreviation = "ECS"
    _description = "(un)satisfiable iff ~C is (un)satisfiable."


class StatusCounterSatisfiable(StatusEquiCounterSatisfiable):
    _name = "CounterSatisfiable"
    _abbreviation = "CSA"
    _description = "- Possible dataforms are Models of Ax | ~C."


class StatusFinitelyCounterSatisfiable(StatusCounterSatisfiable):
    _name = "FinitelyCounterSatisfiable"
    _abbreviation = "FCS"
    _description = "- Possible dataforms are FiniteModels of Ax | ~C."


class StatusSatisfiableCounterTheorem(StatusCounterTheorem, StatusTheorem):
    _name = "SatisfiableCounterTheorem"
    _abbreviation = "SCT"
    _description = "- Possible dataforms are Models of Ax with Proofs of ~C from Ax."


class StatusCounterEquivalent(
    StatusSatisfiableCounterTheorem, StatusCounterSatisfiable
):
    _name = "CounterEquivalent"
    _abbreviation = "CEQ"
    _description = "- Possible dataforms are Proofs of ~C from Ax and of Ax from ~C."


class StatusUnsatisfiableConclusion(
    StatusSatisfiableCounterTheorem, StatusCounterSatisfiable
):
    _name = "UnsatisfiableConclusion"
    _abbreviation = "UNC"
    _description = "- Possible dataforms are Proofs of ~C."


class StatusWeakerCounterConclusion(
    StatusSatisfiableCounterTheorem, StatusCounterSatisfiable
):
    _name = "WeakerCounterConclusion"
    _abbreviation = "WCC"
    _description = "- See CounterTheorem and CounterSatisfiable."


class StatusFinitelyUnsatisfiable(Status):
    _name = "FinitelyUnsatisfiable"
    _abbreviation = "FUN"
    _description = "(i.e., no finite interpretations are finite models of C)."


class StatusEquivalentCounterTheorem(
    StatusCounterEquivalent, StatusFinitelyUnsatisfiable
):
    _name = "EquivalentCounterTheorem"
    _abbreviation = "ECT"
    _description = "- See CounterEquivalent."


class StatusUnsatisfiable(
    StatusFinitelyUnsatisfiable, StatusCounterEquivalent, StatusUnsatisfiableConclusion
):
    _name = "Unsatisfiable"
    _abbreviation = "UNS"
    _description = (
        "- Possible dataforms are Proofs of Ax and of C, and Refutations of F."
    )


class StatusWeakerUnsatisfiableConclusion(
    StatusWeakerCounterConclusion, StatusUnsatisfiableConclusion
):
    _name = "WeakerUnsatisfiableConclusion"
    _abbreviation = "WUC"
    _description = "- See Unsatisfiable and WeakerCounterConclusion."


class StatusWeakerCounterTheorem(StatusWeakerCounterConclusion):
    _name = "WeakerCounterTheorem"
    _abbreviation = "WCT"
    _description = "- See CounterSatisfiable."


class StatusUnsatisfiableConclusionContradictoryAxioms(
    StatusSatisfiableCounterConclusionContradictoryAxioms
):
    _name = "UnsatisfiableConclusionContradictoryAxioms"
    _abbreviation = "UCA"
    _description = "- SatisfiableCounterConclusionContradictoryAxioms."


class StatusNoConsequence(StatusSatisfiable, StatusCounterSatisfiable):
    _name = "NoConsequence"
    _abbreviation = "NOC"
    _description = "of Ax | ~C."


class StatusNoSuccess(Status):
    _name = "NoSuccess"
    _abbreviation = "NOS"
    _description = "The logical data has not been processed successfully (yet)."


class StatusOpen(StatusNoSuccess):
    _name = "Open"
    _abbreviation = "OPN"
    _description = "A success value has never been established."


class StatusUnknown(StatusNoSuccess):
    _name = "Unknown"
    _abbreviation = "UNK"
    _description = "Success value unknown, and no assumption has been made."


class StatusStopped(StatusUnknown):
    _name = "Stopped"
    _abbreviation = "STP"
    _description = (
        "Software attempted to process the data, and stopped without a success status."
    )


class StatusError(StatusStopped):
    _name = "Error"
    _abbreviation = "ERR"
    _description = "Software stopped due to an error."


class StatusOSError(StatusError):
    _name = "OSError"
    _abbreviation = "OSE"
    _description = "Software stopped due to an operating system error."


class StatusInputError(StatusError):
    _name = "InputError"
    _abbreviation = "INE"
    _description = "Software stopped due to an input error."


class StatusUsageError(StatusInputError):
    _name = "UsageError"
    _abbreviation = "USE"
    _description = "Software stopped due to an ATP system usage error."


class StatusSyntaxError(StatusInputError):
    _name = "SyntaxError"
    _abbreviation = "SYE"
    _description = "Software stopped due to an input syntax error."


class StatusSemanticError(StatusInputError):
    _name = "SemanticError"
    _abbreviation = "SEE"
    _description = "Software stopped due to an input semantic error."


class StatusTypeError(StatusSemanticError):
    _name = "TypeError"
    _abbreviation = "TYE"
    _description = (
        "Software stopped due to an input type error (for typed logical data)."
    )


class StatusForced(StatusStopped):
    _name = "Forced"
    _abbreviation = "FOR"
    _description = "Software was forced to stop by an external force."


class StatusUser(StatusForced):
    _name = "User"
    _abbreviation = "USR"
    _description = "Software was forced to stop by the user."


class StatusGaveUp(StatusStopped):
    _name = "GaveUp"
    _abbreviation = "GUP"
    _description = "Software gave up of its own accord."


class StatusResourceOut(StatusForced, StatusGaveUp):
    _name = "ResourceOut"
    _abbreviation = "RSO"
    _description = "Software stopped because some resource ran out."


class StatusTimeout(StatusResourceOut):
    _name = "Timeout"
    _abbreviation = "TMO"
    _description = "Software stopped because the CPU time limit ran out."


class StatusMemoryOut(StatusResourceOut):
    _name = "MemoryOut"
    _abbreviation = "MMO"
    _description = "Software stopped because the memory limit ran out."


class StatusIncomplete(StatusGaveUp):
    _name = "Incomplete"
    _abbreviation = "INC"
    _description = "Software gave up because it's incomplete."


class StatusNotTried(StatusUnknown):
    _name = "NotTried"
    _abbreviation = "NTT"
    _description = "Software has not tried to process the data."


class StatusInappropriate(StatusGaveUp, StatusNotTried):
    _name = "Inappropriate"
    _abbreviation = "IAP"
    _description = "Software gave up because it cannot process this type of data."


class StatusInProgress(StatusUnknown):
    _name = "InProgress"
    _abbreviation = "INP"
    _description = "Software is still running."


class StatusNotTriedYet(StatusNotTried):
    _name = "NotTriedYet"
    _abbreviation = "NTY"
    _description = (
        "Software has not tried to process the data yet, but might in the future."
    )



from gavel.dialects.tptp.parser import (
    TPTPParser,
    TPTPProblemParser,
    SimpleTPTPProofParser,
)
from gavel.logic import logic
from gavel.logic import problem
from gavel.logic import solution
from gavel.logic import sources

from ..test_base.test_parser import (
    TestLogicParser,
    check_wrapper,
    TestProblemParser,
    TestProofParser,
)


class TestTPTPProblemParser(TestProblemParser):
    _parser_cls = TPTPParser



from gavel.dialects.base.dialect import Dialect
from gavel.dialects.tptp.compiler import TPTPCompiler
from gavel.dialects.tptp.parser import SimpleTPTPProofParser
from gavel.dialects.tptp.parser import TPTPProblemParser
from gavel.logic.problem import Problem


class TPTPProblemDialect(Dialect):
    _parser_cls = TPTPProblemParser
    _compiler_cls = TPTPCompiler

    @classmethod
    def _identifier(self):
        return "tptp"


# For legacy code
TPTPDialect = TPTPProblemDialect


class TPTPProofDialect(Dialect):
    _parser_cls = SimpleTPTPProofParser
    _compiler_cls = TPTPCompiler

    @classmethod
    def _identifier(cls) -> str:
        return "tptp-proof"



import unittest
from gavel.logic import logic, logic_utils


class TestFormulaUtils(unittest.TestCase):
    def test_get_vars_in_formula(self):
        formula = logic.NaryFormula(
            logic.BinaryConnective.CONJUNCTION,
            [logic.PredicateExpression("c", [logic.Variable("x")]),
             logic.PredicateExpression("n", [logic.Constant("atom2")]),
             logic.BinaryFormula(logic.PredicateExpression("h", [logic.Constant("k"), logic.Variable("z")]),
                                 logic.BinaryConnective.DISJUNCTION,
                                 logic.PredicateExpression("c", [logic.Variable("x")]))])
        variables = logic_utils.get_vars_in_formula(formula)
        self.assertTrue(logic.Variable("x") in variables)
        self.assertFalse(logic.Variable("k") in variables)  # k is a constant
        self.assertTrue(logic.Variable("z") in variables)

    def test_flatten_formula(self):
        # ((abc & def) & ((ghi & (abc & def) & (ghi | jkl))))
        formula = logic.BinaryFormula(
            logic.BinaryFormula(
                logic.PredicateExpression("abc", [logic.Variable("x")]),
                logic.BinaryConnective.CONJUNCTION,
                logic.PredicateExpression("def", [logic.Variable("y")])
            ),
            logic.BinaryConnective.CONJUNCTION,
            logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, [
                logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, [
                    logic.PredicateExpression("ghi", [logic.Constant("a")])
                ]),
                logic.BinaryFormula(
                    logic.PredicateExpression("abc", [logic.Variable("x")]),
                    logic.BinaryConnective.CONJUNCTION,
                    logic.PredicateExpression("def", [logic.Variable("y")])
                ),
                logic.NaryFormula(logic.BinaryConnective.DISJUNCTION, [
                    logic.PredicateExpression("ghi", [logic.Constant("a")]),
                    logic.PredicateExpression("jkl", [logic.Constant("b")])
                ])
            ])
        )
        # result should be a n-ary formula
        self.assertTrue(isinstance(logic_utils.binary_to_nary(formula, logic.BinaryConnective.CONJUNCTION), logic.NaryFormula))
        # formula should not have conjunctions in layer below
        self.assertFalse(any(isinstance(f, logic.BinaryFormula)
                             or (isinstance(f, logic.NaryFormula) and f.operator == logic.BinaryConnective.CONJUNCTION)
                             for f in logic_utils.binary_to_nary(formula, logic.BinaryConnective.CONJUNCTION).formulae))
        # don't flatten disjunction
        self.assertTrue(any(isinstance(f, logic.NaryFormula) and f.operator == logic.BinaryConnective.DISJUNCTION
                        for f in logic_utils.binary_to_nary(formula, logic.BinaryConnective.CONJUNCTION).formulae))

    def test_substitute_var_in_formula(self):
        for formula in [logic.PredicateExpression("abc", [logic.Variable("X"), logic.Variable("Y")]),
                        logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, [
                            logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, [
                                logic.PredicateExpression("ghi", [logic.Variable("X")])
                            ]),
                            logic.BinaryFormula(
                                logic.PredicateExpression("abc", [logic.Variable("X")]),
                                logic.BinaryConnective.CONJUNCTION,
                                logic.PredicateExpression("def", [logic.Variable("Y")])
                            ),
                            logic.NaryFormula(logic.BinaryConnective.DISJUNCTION, [
                                logic.PredicateExpression("ghi", [logic.Constant("a")]),
                                logic.PredicateExpression("jkl", [logic.Variable("X")])
                            ])
                        ])]:
            new_constant = logic.Constant("c1")
            substituted = logic_utils.substitute_var_in_formula(formula, logic.Variable("X"), new_constant)
            self.assertFalse(logic.Variable("X") in logic_utils.get_vars_in_formula(substituted))
            self.assertTrue("c1" in substituted.symbols())
            (self.assertTrue(logic.Variable("Y") in logic_utils.get_vars_in_formula(substituted),
                             f"Found only variables "
                             f"{[str(var) for var in logic_utils.get_vars_in_formula(substituted)]}"
                             f" in formula {substituted}"))

    def test_operator_overloading(self):
        formula = logic.PredicateExpression("abc", [logic.Constant("m")])
        negated_formula = ~formula
        self.assertTrue(isinstance(negated_formula, logic.UnaryFormula))
        self.assertEqual(str(negated_formula), str(logic.UnaryFormula(logic.UnaryConnective.NEGATION, formula)))

    def testNNF(self):
        formula_eq = logic.UnaryFormula(
            logic.UnaryConnective.NEGATION,
            logic.BinaryFormula(logic.Variable("x"), logic.BinaryConnective.EQ, logic.Variable("y")))

        formula_biimp = logic.UnaryFormula(
            logic.UnaryConnective.NEGATION,
            logic.BinaryFormula(
                logic.PredicateExpression("weirdly_similar", [logic.Variable("x"), logic.Variable("y")]),
                logic.BinaryConnective.BIIMPLICATION,
                logic.UnaryFormula(
                    logic.UnaryConnective.NEGATION,
                    logic.BinaryFormula(logic.Variable("x"), logic.BinaryConnective.EQ, logic.Variable("y"))),
            )
        )

        formula_combination = logic.UnaryFormula(
            logic.UnaryConnective.NEGATION,
            logic.QuantifiedFormula(
                logic.Quantifier.UNIVERSAL,
                [logic.Variable("x"), logic.Variable("y")],
                logic.BinaryFormula(
                    logic.PredicateExpression("weirdly_similar", [logic.Variable("x"), logic.Variable("y")]),
                    logic.BinaryConnective.BIIMPLICATION,
                    logic.BinaryFormula(
                        logic.UnaryFormula(
                            logic.UnaryConnective.NEGATION,
                            logic.BinaryFormula(logic.Variable("x"), logic.BinaryConnective.EQ, logic.Variable("y"))),
                        logic.BinaryConnective.CONJUNCTION,
                        logic.BinaryFormula(
                            logic.BinaryFormula(
                                logic.PredicateExpression("weird", [logic.Variable("x")]),
                                logic.BinaryConnective.DISJUNCTION,
                                logic.UnaryFormula(
                                    logic.UnaryConnective.NEGATION,
                                    logic.PredicateExpression("normal", [logic.Variable("x")]))
                            ),
                            logic.BinaryConnective.IMPLICATION,
                            logic.PredicateExpression("weird", [logic.Variable("y")]),
                        )
                    )
                )
            )
        )

        for formula in [formula_eq, formula_biimp, formula_combination]:
            print("Formula:", formula)
            formula = logic_utils.convert_to_nnf(formula)
            print("Formula in NNF:", formula)
            # weak assertion, does not test correct propagation of negation
            self.assertNotIsInstance(formula, logic.UnaryFormula)

    def testCNF(self):
        formula = logic.BinaryFormula(
            logic.PredicateExpression("weirdly_similar", [logic.Variable("x"), logic.Variable("y")]),
            logic.BinaryConnective.DISJUNCTION,
            logic.NaryFormula(logic.BinaryConnective.CONJUNCTION, [
                logic.UnaryFormula(logic.UnaryConnective.NEGATION,
                                   logic.BinaryFormula(logic.Variable("x"), logic.BinaryConnective.EQ,
                                                       logic.Variable("y"))),
                logic.BinaryFormula(
                    logic.PredicateExpression("weird", [logic.Variable("x")]),
                    logic.BinaryConnective.DISJUNCTION,
                    logic.UnaryFormula(
                        logic.UnaryConnective.NEGATION,
                        logic.PredicateExpression("normal", [logic.Variable("x")]))
                ),
                logic.PredicateExpression("weird", [logic.Variable("y")]),
            ])
        )
        print("Formula:", formula)
        formula = logic_utils.convert_to_cnf(formula)
        print("Formula in CNF:", formula)
        self.assertIsInstance(formula, logic.NaryFormula)
        self.assertEqual(formula.operator, logic.BinaryConnective.CONJUNCTION)
        for clause in formula.formulae:
            self.assertIsInstance(clause, logic.NaryFormula)
            self.assertEqual(clause.operator, logic.BinaryConnective.DISJUNCTION)
            for literal in clause.formulae:
                self.assertTrue(logic_utils.is_literal(literal))


if __name__ == '__main__':
    unittest.main()



PROVERS = {}


def register_prover(name):
    """
    A class wrapper that registers a class under `name`. Thereafter this
    prover is known for the remaining execution and can be accessed via
    `get_prover` or in the commandline interface using its name.
    :param name: Identifier for this prover
    """

    def register_class(cls):
        PROVERS[name] = cls
        return cls

    return register_class


def get_prover(name):
    """
    Returns a registered prover identified by `name`.
    :param name:  Identifier of some prover
    :return: The prover associated with `name`
    """
    return PROVERS[name]


__all__ = ["get_prover", "register_prover"]



from unittest import TestCase
from gavel.logic import logic
from gavel.prover.registry import register_prover, get_prover
from gavel.prover.base.interface import BaseProverInterface
from gavel.logic.solution import Proof, ProofStep


def simple_prover(problem):
    for premise in problem.premises:
        if premise.formula == problem.conjecture.formula:
            # Create a proof structure
            p = Proof(steps=[ProofStep(formula=premise.formula)])
            return p
    return None


@register_prover("test")
class DummyTestProver(BaseProverInterface):
    def _submit_problem(self, problem_instance, *args, **kwargs):
        return Proof(
            premises=problem_instance.premises,
            steps=[ProofStep(formula=logic.DefinedConstant.FALSUM, name="name")],
        )


class TestSimpleProver(TestCase):
    def test_registry(self):
        self.assertEqual(get_prover("test"), DummyTestProver)
